package com.kob.backend.consumer.utils;

import com.alibaba.fastjson2.JSONObject;
import com.kob.backend.consumer.WebSocketServer;
import com.kob.backend.pojo.Bot;
import com.kob.backend.pojo.Record;
import com.kob.backend.pojo.User;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;

import java.util.*;
import java.util.concurrent.locks.ReentrantLock;

// 用来管理整个游戏流程:这里需要多线程
public class Game extends Thread {
    // 游戏地图:行数,列数,内部障碍物数量
    private final Integer rows;
    private final Integer cols;
    private final Integer inner_walls_count;
    // 地图数组
    private final int[][] g;
    // 定义"上右下左"四个方向的 dx, dy偏移量
    private final int[] dx = {-1, 0, 1, 0}, dy = {0, 1, 0, -1};
    private final Player playerA, playerB;
    // 玩家下一步操作状态定义
    private Integer nextStepA = null;
    private Integer nextStepB = null;
    // 加锁:解决读写冲突用
    private final ReentrantLock lock = new ReentrantLock();
    // 整个游戏的当前状态: playing(正在进行) --> finished(游戏结束)
    private String status = "playing";
    // 定义失败者: all(平局), A(A输), B(B输)
    private String loser = "";
    // 传送消息给 BotRunningSystem 的链接
    private static final String addBotUrl = "http://127.0.0.1:3002/bot/add/";


    // 初始化(有参)构造函数
    public Game(Integer rows, Integer cols, Integer inner_walls_count, Integer idA, Bot botA, Integer idB, Bot botB) {
        this.rows = rows;
        this.cols = cols;
        this.inner_walls_count = inner_walls_count;
        this.g = new int[rows][cols];

        // 定义默认 botId 值: botId 等于 -1 时表示玩家"亲自出战",非负则表示由对应 id 的 bot 操作(Ai)
        Integer botIdA = -1, botIdB = -1;
        // 定义默认 bot 代码值
        String botCodeA = "", botCodeB = "";
        // 判断 botA 不等于 null 时,对 botA 的 id 和 code 重新赋值;否则取默认值
        if (botA != null) {
            botIdA = botA.getId();
            // botA.getContent() 是 bot 的代码区域
            botCodeA = botA.getContent();
        }
        // 判断 botB 不等于 null 时,对 botB 的 id 和 code 重新赋值
        if (botB != null) {
            botIdB = botB.getId();
            botCodeB = botB.getContent();
        }

        playerA = new Player(idA, botIdA, botCodeA, rows - 2, 1, new ArrayList<>());
        playerB = new Player(idB, botIdB, botCodeB, 1, cols - 2, new ArrayList<>());
    }

    public Player getPlayerA() {
        return playerA;
    }

    public Player getPlayerB() {
        return playerB;
    }

    public int getRows() {
        return rows;
    }

    public int getCols() {
        return cols;
    }

    public int getInnerWallsCount() {
        return inner_walls_count;
    }

    // 将地图信息转换成字符串
    private String getGameMapString() {
        StringBuilder res = new StringBuilder();
        // 将地图数据展开成一维
        /*
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                res.append(g[i][j]);
            }
        }
        */
        for (int[] row : g) {
            for (int col : row) {
                res.append(col);
            }
        }

        return res.toString();
    }

    public void setNextStepA(Integer nextStepA) {
        // 为了防止两方读写冲突,需要加线程锁之后操作 nextStep 值
        lock.lock();
        try {
            this.nextStepA = nextStepA;
        } finally {
            // 操作完成后解锁
            lock.unlock();
        }
    }

    public void setNextStepB(Integer nextStepB) {
        lock.lock();
        try {
            this.nextStepB = nextStepB;
        } finally {
            lock.unlock();
        }
    }

    // 返回生成的地图
    public int[][] getG() {
        return g;
    }

    // 联通检测方法---true(联通)---false(不通),参数: 起点坐标 sx,sy ,终点坐标 tx,ty
    private boolean check_connectivity(int sx, int sy, int tx, int ty) {
        // 起点就是终点时,结果联通,直接返回 true
        if (sx == tx && sy == ty) return true;
        g[sx][sy] = 1;

        //枚举"上右下左"四个方向,求当前点下一个相邻点的坐标
        for (int i = 0; i < 4; i++) {
            int x = sx + dx[i];
            int y = sy + dy[i];

            // 判断是否撞到障碍物( g[x][y] == 0 表示没有碰到障碍物 ),如果没有赚到障碍物,且可以找到重点的话,返回 true(联通)
            if (x >= 0 && x < this.rows && y >= 0 && y < this.cols && g[x][y] == 0) {
                if (check_connectivity(x, y, tx, ty)) {
                    // 还原状态
                    g[sx][sy] = 0;
                    return true;
                }
            }
        }

        // 还原状态
        g[sx][sy] = 0;
        return false;
    }

    // 画地图
    public boolean draw() {
        // 一开始现将所有障碍物初始化为 false
        for (int r = 0; r < this.rows; r++) {
            for (int c = 0; c < this.cols; c++) {
                // 0 表示空地, 1 表示障碍物(墙)
                g[r][c] = 0;
            }
        }

        // 地图左右边缘障碍物
        for (int r = 0; r < this.rows; r++) {
            g[r][0] = g[r][this.cols - 1] = 1;
        }
        // 地图上下边缘障碍物
        for (int c = 0; c < this.cols; c++) {
            g[0][c] = g[this.rows - 1][c] = 1;
        }

        Random random = new Random();
        // 创建内部随机障碍物,每次计算时会生成两个障碍物,因此这里的循环次数 this.inner_walls_count 需要处以 2
        for (int i = 0; i < this.inner_walls_count / 2; i++) {
            // 避免位置重复:重复遍历 1000 次,只要找到了已经存在障碍物的位置就禁止随机
            for (int j = 0; j < 1000; j++) {
                // 找出本次随机到的行-r 列-c 坐标
                int r = random.nextInt(this.rows);  // random.nextInt(7) 返回 0-7之间的随机整数
                int c = random.nextInt(this.cols);

                // 中心对称:当本坐标或者它的中心对称坐标已经存在障碍物了,则重新计算下一个坐标
                if (g[r][c] == 1 || g[this.rows - 1 - r][this.cols - 1 - c] == 1)
                    continue;

                // 避免内部障碍物覆盖掉左下角和右上角的 A-B 角色出发点,如果是这两个位置,则重新计算下一个坐标
                if (r == this.rows - 2 && c == 1 || r == 1 && c == this.cols - 2)
                    continue;

                // 将计算求得的随机障碍物合法的位置赋值为 1 ,以对该位置进行绘制(包括本坐标及其中心对称坐标)
                g[r][c] = g[this.rows - 1 - r][this.cols - 1 - c] = 1;

                // 1000 次遍历中,规定数量的内部障碍物已经够数之后就 break 掉
                break;
            }
        }

        // 确保 A-B 角色的运动区域是联通的:如果不连通,则直接在创建地图对象之前取消绘制( return false )
        return check_connectivity(this.rows - 2, 1, 1, this.cols - 2);
    }

    public void createMap() {
        // 循环绘制:如果发现哪次循环中画地图成功了,则跳出循环,绘制结束
        for (int i = 0; i < 1000; i++) {
            if (draw())
                break;
        }
    }

    // 将当前的局面信息编码成字符串
    private String getInput(Player player) {
        Player me, you;
        if (playerA.getId().equals(player.getId())) {
            me = playerA;
            you = playerB;
        } else {
            me = playerB;
            you = playerA;
        }

        // 地图#me.sx#me.sy#我的操作#you.sx#you.sy#对手操作
        // getStepsString() 操作序列可能是空的,为了防止分割出错,在这里加上()
        return getGameMapString() + "#"
                + me.getSx() + "#"
                + me.getSy() + "#("
                + me.getStepsString() + ")#"
                + you.getSx() + "#"
                + you.getSy() + "#("
                + you.getStepsString() + ")";
    }

    // 发送 bot 代码让服务器自动执行
    private void sendBotCode(Player player) {
        //// botId 等于 -1 时表示玩家"亲自出战",非负则表示由对应 id 的 bot 操作(Ai)
        if (player.getBotId() == -1) return;    // 当人来亲自操作时,直接 return, 不需要执行代码

        // 不是人来控制时(改为 Bot 按照代码逻辑自动执行)
        // 封装数据包
        MultiValueMap<String, String> data = new LinkedMultiValueMap<>();
        data.add("user_id", player.getId().toString());
        data.add("bot_code", player.getBotCode());
        data.add("input", getInput(player));

        // 向 BotRunningSystem 发送消息
        WebSocketServer.restTemplate.postForObject(addBotUrl, data, String.class);
    }

    // 获取两名玩家的下一步操作
    private boolean nextStep() {
        try {
            // 返回下一步操作之前,先睡眠 200 毫秒,用于等待前端渲染完成,避免渲染速度跟不上后端运算处理速度
            // 前端定义了每 1 秒钟 5 个格子,则每 200 毫秒走一格(每一步至少需要 200 毫秒的时间)
            Thread.sleep(200);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }

        sendBotCode(playerA);
        sendBotCode(playerB);


        // 循环 50 秒钟,每次 100 毫秒,判断用户输入(50 * 100 = 5000 ms后如果没有接收到输入,则判断输赢)
        for (int i = 0; i < 50; i++) {
            try {
                // 两次接收输入的等待间隔
                Thread.sleep(100);
                lock.lock();
                try {
                    // 判断两名玩家是否都已经读取到输入:如果读入都非空,则表示读取结束,返回 true
                    if (nextStepA != null && nextStepB != null) {
                        // 将下一步操作存下来
                        playerA.getSteps().add(nextStepA);
                        playerB.getSteps().add(nextStepB);

                        // 成功获取到两名玩家的下一步操作,返回 true
                        return true;
                    }
                } finally {
                    lock.unlock();
                }
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        return false;
    }

    // 碰撞合法性检测辅助函数:需要判断 snakeA 和 snakeB 是否重合
    private boolean checkValid(List<Cell> cellsA, List<Cell> cellsB) {
        int n = cellsA.size();  // cellsA 的当前长度
        Cell cell = cellsA.get(n - 1);   // 取出 cellsA 的蛇头位置(数组最后一位)(第0个元素是蛇尾巴)
        if (g[cell.x][cell.y] == 1) return false;    // 判断 A 的最后一位(蛇头)的坐标是否是障碍物(值为1),如果是障碍物则返回false(判断已碰撞)

        // 判断 A蛇头位置是否和A除了蛇头以外的其他身体部分坐标有重合
        for (int i = 0; i < n - 1; i++) {
            if (cellsA.get(i).x == cell.x && cellsA.get(i).y == cell.y)
                return false;
        }
        for (Cell cellB : cellsB) {   // 判断 A蛇头位置是否和B蛇身体各部分坐标有重合:有重合返回不合法(表示已经碰撞)
            if (cellB.x == cell.x && cellB.y == cell.y)
                return false;
        }

        // 以上判断条件如果都合法,则返回 true合法(表示没有碰撞)
        return true;
    }

    // 判断两名玩家下一步操作是否合法
    private void judge() {
        // 取出 A,B 两蛇
        List<Cell> cellsA = playerA.getCells();
        List<Cell> cellsB = playerB.getCells();

        boolean validA = checkValid(cellsA, cellsB);
        boolean validB = checkValid(cellsB, cellsA);

        // 如果 A 和 B 有人这一步不合法(已经产生碰撞),将游戏状态改为 "finished",表示游戏结束
        if (!validA || !validB) {
            status = "finished";

            // 游戏结束时,判断输赢
            if (!validA && !validB) {
                loser = "all";
            } else if (!validA) {
                loser = "A";
            } else {
                loser = "B";
            }
        }
    }

    // 向两个 Client 广播信息
    private void sendAllMessage(String message) {
        if (WebSocketServer.users.get(playerA.getId()) != null)
            WebSocketServer.users.get(playerA.getId()).sendMessage(message);
        if (WebSocketServer.users.get(playerB.getId()) != null)
            WebSocketServer.users.get(playerB.getId()).sendMessage(message);
    }

    // 向两个 Client 传递移动信息
    private void sendMove() {
        //由于这里需要读入 nextStepA, nextStepB, 这里需要加一下线程锁
        lock.lock();
        try {
            JSONObject resp = new JSONObject();
            resp.put("event", "move");   // event(事件类型): move(移动信息)
            resp.put("a_direction", nextStepA);  // a_direction( playerA 移动的方向)
            resp.put("b_direction", nextStepB);  // b_direction( playerB 移动的方向)
            sendAllMessage(resp.toJSONString());
            // 本次获取完下一步操作的同时需要进行再下一步操作输入,在这之前,需要将当前的 nextStep 清空
            nextStepA = nextStepB = null;
        } finally {
            lock.unlock();
        }
    }

    // 更新玩家积分
    private void updateUserRating(Player player, Integer rating) {
        User user = WebSocketServer.userMapper.selectById(player.getId());
        user.setRating(rating); // 修改用户 rating 信息
        WebSocketServer.userMapper.updateById(user);    // 更新数据库的 rating 信息
    }


    // 将游戏结果存到数据库中
    private void saveToDataBase() {
        // 取出用户天梯积分并更新
        Integer ratingA = WebSocketServer.userMapper.selectById(playerA.getId()).getRating();
        Integer ratingB = WebSocketServer.userMapper.selectById(playerB.getId()).getRating();
        if ("A".equals(loser)) {    // A 是 Loser 则: 玩家A积分-2,同时玩家B积分+5
            ratingA -= 2;
            ratingB += 5;
        } else if ("B".equals(loser)) {
            ratingA += 5;
            ratingB -= 2;
        }
        updateUserRating(playerA, ratingA);
        updateUserRating(playerB, ratingB);


        Record record = new Record(
                null,
                playerA.getId(),
                playerA.getSx(),
                playerA.getSy(),
                playerB.getId(),
                playerB.getSx(),
                playerB.getSy(),
                playerA.getStepsString(),
                playerB.getStepsString(),
                getGameMapString(),
                loser,
                new Date()
        );

        WebSocketServer.recordMapper.insert(record);
    }


    // 向两个 Client 公布结果
    private void sendResult() {
        lock.lock();
        try {
            JSONObject resp = new JSONObject();
            resp.put("event", "result"); // event(事件类型): result(结果)
            resp.put("loser", loser);    // 失败者
            // 将最后碰撞时的蛇的眼睛指向传给前端
            resp.put("a_eyes_finally_direction", nextStepA);
            resp.put("b_eyes_finally_direction", nextStepB);
            // 发送结果之前,先将结果存到数据库中
            saveToDataBase();
            sendAllMessage(resp.toJSONString());
        } finally {
            lock.unlock();
        }
    }

    // 重写线程函数
    @Override
    public void run() {
        // 循环 1000 步: 13格x14格x(最大)3步/格 < 1000，保证 1000 次一定可以走完
        for (int i = 0; i < 1000; i++) {
            if (nextStep()) { // 是否获取到两条蛇的下一步操作
                // 判断下一步操作是否合法
                judge();

                // 如果游戏状态还是 playing,则向两个玩家广播移动信息
                if ("playing".equals(status)) {
                    sendMove();
                } else if ("finished".equals(status)) {
                    // 如果游戏状态已经更改为了 finished,则将结果返回给 Client 并 break 中断执行后续操作
                    sendResult();
                    break;
                }

            } else {
                // 未获取到某蛇的下一步操作,则将游戏状态改为 finished
                status = "finished";
                // 判断失败者:这里需要加锁,因为涉及到对 nextStep 的读操作
                lock.lock();
                try {
                    if (nextStepA == null && nextStepB == null) {
                        loser = "all";
                    } else if (nextStepA == null) {
                        loser = "A";
                    } else {
                        loser = "B";
                    }
                } finally {
                    lock.unlock();
                }
            }
        }
    }
}
